from django.conf.urls import url
from django.urls import include, path
from django.views.generic.base import RedirectView

from ..core.apps.groups.urls import urlpatterns as groups_urls
from ..core.apps.users.urls import urlpatterns as users_urls
from ..core.apps.users.urls import extra_urlpatterns as users_extra_urls
from ..core.apps.profiles.urls import base_urlpatterns as profiles_urls
from ..core.apps.profiles.urls import extra_urlpatterns as profiles_extra_urls
from ..core import app_settings

from .views import AppLoginView, AppSignupView, AppLogoutView


django_sso_app_profile_urlpatterns = [
    path('profile/', RedirectView.as_view(url=app_settings.REMOTE_PROFILE_VIEW_URL, permanent=False),
         name='profile'),
    path('profile/update/', RedirectView.as_view(url=app_settings.REMOTE_PROFILE_UPDATE_URL, permanent=False),
         name='profile.update'),
    path('profile/complete/', RedirectView.as_view(url=app_settings.REMOTE_PROFILE_COMPLETE_URL, permanent=False),
         name='profile.complete'),
]

django_sso_app_i18n_urlpatterns = django_sso_app_profile_urlpatterns

django_sso_app_urlpatterns = [
    url(r'^logout/$', AppLogoutView.as_view(), name='account_logout'),
    url(r'^login/$', AppLoginView.as_view(), name='account_login'),
    url(r'^signup/$', AppSignupView.as_view(), name='account_signup'),
    url(r'^email/$', RedirectView.as_view(url=app_settings.REMOTE_EMAIL_URL, permanent=False), name='account_email'),

    url(r"^password/change/$", RedirectView.as_view(url=app_settings.BACKEND_URL + '/password/change/', permanent=False),
        name='account_change_password'),
    url(r"^password/reset/$", RedirectView.as_view(url=app_settings.BACKEND_URL + '/password/reset/', permanent=False),
        name='account_reset_password'),
    url(r"^email/$", RedirectView.as_view(url=app_settings.BACKEND_URL + '/email/', permanent=False),
        name='account_email'),
]

django_sso_app_api_urlpatterns = [
    url(r'^api/v1/auth/groups/', include(groups_urls)),
    url(r'^api/v1/auth/users/', include(users_urls)),
    url(r'^api/v1/auth/profiles/', include(profiles_urls)),
] + [
    url(r'^api/v1/auth/', include(users_extra_urls)),
    url(r'^api/v1/auth/', include(profiles_extra_urls)),
]



# Importing core models in order to enable "dumpdata"

from .core.apps.devices.models import *
from .core.apps.groups.models import *
from .core.apps.passepartout.models import *
from .core.apps.profiles.models import *
from .core.apps.services.models import *
from .core.apps.status.models import *
from .core.apps.events.models import *



from ...serializers import AbsoluteUrlSerializer
from .models import EmailAddress


class EmailSerializer(AbsoluteUrlSerializer):
    class Meta:
        model = EmailAddress
        read_only_fields = ('email', 'verified', 'primary')
        fields = read_only_fields



from django.utils.translation import gettext_lazy as _

LANGUAGES = (
    ## Customize this
    ('it', _('it')),
    ('en', _('en')),
    ('es', _('es')),
    ('pt', _('pt')),
    ('fr', _('fr')),
    ('de', _('de')),
    ('nl', _('nl')),
)



from django.contrib.auth import get_user_model

from django_sso_app.core.tests.factories import UserTestCase
from django_sso_app.core.apps.emails.models import EmailAddress

User = get_user_model()


class TestEmailAddress(UserTestCase):

    def test_user_email_deletes_other_user_emails_when_primary_and_verified(self):
        user = self._get_new_user()

        self.assertEqual(user.emailaddress_set.count(), 1)

        new_email = EmailAddress.objects.create(user=user, email=self._get_random_email())

        self.assertEqual(user.emailaddress_set.count(), 2)

        new_email.primary = True
        new_email.verified = True
        new_email.save()

        self.assertEqual(user.emailaddress_set.count(), 1, 'other user email not confirmed')

    def test_user_can_have_only_one_not_validated_email(self):
        user = self._get_new_user()

        self.assertEqual(user.emailaddress_set.count(), 1)

        _new_email = EmailAddress.objects.create(user=user, email=self._get_random_email(), verified=False)

        self.assertEqual(user.emailaddress_set.count(), 2)

        new_email2 = EmailAddress.objects.create(user=user, email=self._get_random_email(), verified=False)

        self.assertEqual(user.emailaddress_set.count(), 2)

        new_email2.primary = True
        new_email2.verified = True
        new_email2.save()

        self.assertEqual(user.emailaddress_set.count(), 1, 'other user email not confirmed')



from django.conf.urls import url
from django.urls import include, path
from django.contrib.auth.decorators import login_required
#from django.conf import settings
#from django.views.generic.base import RedirectView

from ..core.apps.users.urls import urlpatterns as users_urls
from ..core.apps.users.urls import extra_urlpatterns as users_extra_urls
from ..core.apps.groups.urls import urlpatterns as groups_urls
from ..core.apps.profiles.urls import urlpatterns as profiles_urls
from ..core.apps.profiles.urls import extra_urlpatterns as profiles_extra_urls
from ..core.apps.profiles.views import ProfileView, ProfileUpdateView, ProfileCompleteView
from ..core.apps.services.urls import urlpatterns as services_urls
from ..core.apps.devices.urls import urlpatterns as devices_urls
from ..core.apps.passepartout.urls import urlpatterns as passepartout_urls

from ..core.urls import allauth_urlpatterns, allauth_i18n_urlpatterns
from ..core.api.urls import allauth_api_urlpatterns

django_sso_app_profile_urlpatterns = [
    path('profile/', login_required(ProfileView.as_view()), name='profile'),
    path('profile/update/', login_required(ProfileUpdateView.as_view()), name='profile.update'),
    path('profile/complete/', login_required(ProfileCompleteView.as_view()), name='profile.complete'),
]
django_sso_app_i18n_urlpatterns = django_sso_app_profile_urlpatterns + allauth_i18n_urlpatterns
django_sso_app_urlpatterns = allauth_urlpatterns
django_sso_app_api_urlpatterns = allauth_api_urlpatterns

django_sso_app_api_urlpatterns += [
    url(r'^api/v1/auth/groups/', include(groups_urls)),
    url(r'^api/v1/auth/users/', include(users_urls)),
    url(r'^api/v1/auth/profiles/', include(profiles_urls)),

    url(r'^api/v1/auth/devices/', include(devices_urls)),
    url(r'^api/v1/auth/services/', include(services_urls)),
    url(r'^api/v1/auth/passepartout/', include(passepartout_urls)),

    # url('^api/v1/jwt/refresh/?', TokenRefreshView.as_view(), name='token_refresh'),
    # url('^api/v1/jwt/verify/?', TokenVerifyView.as_view(), name='token_verify'),
] + [
    url(r'^api/v1/auth/', include(users_extra_urls)),
    url(r'^api/v1/auth/', include(profiles_extra_urls)),
]



from django.conf.urls import url
from rest_framework.urlpatterns import format_suffix_patterns

from .views import DeviceApiViewSet


_urlpatterns = [
    url(r'^$', DeviceApiViewSet.as_view({'get': 'list'}), name='rest-list'),
    url(r'^(?P<pk>\d+)/$', DeviceApiViewSet.as_view({'get': 'retrieve', 'delete': 'destroy'}),
        name="rest-detail"),
]

urlpatterns = (format_suffix_patterns(_urlpatterns), 'django_sso_app_device')



from django.contrib import admin

from .models import Passepartout


class PassepartoutAdmin(admin.ModelAdmin):
    search_fields = ('id', 'device__profile__sso_id',
                     'device__profile__user__username',
                     'device__profile__user__email')
    list_display = ('user', 'is_active', 'created_at')


admin.site.register(Passepartout, PassepartoutAdmin)



# Generated by Django 2.2.12 on 2020-07-03 09:39

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_sso_app', '0005_auto_20200519_1725'),
    ]

    operations = [
        migrations.DeleteModel(
            name='EmailAddress',
        ),
    ]



import os
import logging
import datetime
import binascii
from random import seed, randint
import base64

from django.core.management.utils import get_random_secret_key
from django.utils.http import urlencode
from django.http import HttpResponseRedirect
from django.conf import settings

from .tokens.utils import logger
from .exceptions import (ProfileIncompleteException, DectivatedUserException, UnsubscribedUserException,
                         DjangoStaffUsersCanNotLoginException, ServiceSubscriptionRequiredException,
                         AnonymousUserException)
from .permissions import is_authenticated, is_django_staff
from . import app_settings

logger = logging.getLogger('django_sso_app.core')
# initializing seed
SEED = int(base64.b16encode(bytes(get_random_secret_key(), 'utf-8')), 16)
seed(SEED)

JWT_COOKIE_SECURE = not settings.DEBUG


def set_cookie(response, key, value, days_expire=None, secure=JWT_COOKIE_SECURE):
    """
    Sets response auth cookie

    :param response:
    :param key:
    :param value:
    :param days_expire:
    :param secure:
    :return:
    """

    if days_expire is None:
        max_age = app_settings.COOKIE_AGE  # django default
    else:
        max_age = days_expire * 24 * 60 * 60

    expires = datetime.datetime.strftime(datetime.datetime.utcnow() + datetime.timedelta(seconds=max_age),
                                         "%a, %d-%b-%Y %H:%M:%S GMT")

    response.set_cookie(key, value, max_age=None, expires=expires, path='/',
                        domain=app_settings.COOKIE_DOMAIN,
                        secure=secure,
                        httponly=app_settings.COOKIE_HTTPONLY)

    # samesite None cookie patch
    # replaced by "same_site" middleware
    # if app_settings.SAME_SITE_COOKIE_NONE:
    #    response.cookies[key]['samesite'] = 'None'


def invalidate_cookie(response, key):
    """
    Sets response auth cookie to 'Null'

    :param response:
    :param key:
    :return:
    """
    response.set_cookie(key, '', max_age=None, expires='Thu, 01 Jan 1970 00:00:01 GMT', path='/',
                        domain=app_settings.COOKIE_DOMAIN, secure=None, httponly=False)


def set_session_key(request, key, value):
    """
    Sets django sessions request key
    :param request:
    :param key:
    :param value:
    :return:
    """

    #request.session[key] = value

    setattr(request, key, value)


def get_session_key(request, key, default=None):
    """
    Gets django sessions request key
    :param request:
    :param key:
    :param default:
    :return:
    """

    #return request.session.get(key, default)

    return getattr(request, key, default)


def redirect_to_profile_complete(user):
    if app_settings.BACKEND_ENABLED:
        url = app_settings.PROFILE_COMPLETE_URL
    else:
        _qs = urlencode({'next': app_settings.SERVICE_URL})
        url = '{}{}?{}'.format(app_settings.BACKEND_URL, app_settings.PROFILE_COMPLETE_URL, _qs)

    logger.info('User {} must complete profile,'
                ' redirecting to {} ...'
                .format(user, url))

    response = HttpResponseRedirect(redirect_to=url)

    raise ProfileIncompleteException(response)


def redirect_to_service_subscription(user, service_url=None):
    if app_settings.BACKEND_ENABLED:
        url = app_settings.LOGIN_URL
    else:
        if service_url is None:
            service_url = app_settings.SERVICE_URL

        _qs = urlencode({'next': service_url})
        url = '{}{}?{}'.format(app_settings.BACKEND_URL, app_settings.LOGIN_URL, _qs)

    logger.info('User {} must agree to the Terms of Service,'
                ' redirecting to {} ...'
                .format(user, url))

    response = HttpResponseRedirect(redirect_to=url)

    raise ServiceSubscriptionRequiredException(response)


def check_user_can_login(user, skip_profile_completion_checks=False, skip_service_subscription_checks=False):
    """
    Check user login ability
    :param user:
    :param skip_profile_completion_checks:
    :param skip_service_subscription_checks:
    :return:
    """
    if is_authenticated(user):
        if is_django_staff(user):
            raise DjangoStaffUsersCanNotLoginException()
        elif not user.is_active:
            raise DectivatedUserException()
        else:
            if user.sso_app_profile.is_unsubscribed:
                raise UnsubscribedUserException()
            else:
                if not skip_profile_completion_checks:
                    if user.sso_app_profile.is_incomplete:
                        redirect_to_profile_complete(user)

                if not skip_service_subscription_checks:
                    if user.sso_app_profile.must_subscribe:
                        redirect_to_service_subscription(user)
    else:
        raise AnonymousUserException()


def get_random_token(length=4):
    """
    Returns random token
    :param length:
    :return:
    """
    start = randint(0, 4095 - length)
    end = start + length

    return binascii.hexlify(os.urandom(2048)).decode("utf-8")[start:end]


def get_random_fingerprint(request):
    """
    Returns random fingerprint
    :param request:
    :return:
    """
    return 'random_{}_{}'.format(id(request), get_random_token(8))



import logging

from django.contrib.auth import get_user_model
from django.http import Http404
from rest_framework import mixins
from rest_framework import permissions
from rest_framework import status
from rest_framework import viewsets
from rest_framework.response import Response

from .models import Service, Subscription
from .serializers import SubscriptionSerializer, ServiceSerializer, ServiceSubscriptionSerializer
from ..profiles.models import Profile
from ...permissions import is_staff, is_authenticated
from ...tokens.utils import get_request_jwt
from ..devices.utils import renew_response_jwt

logger = logging.getLogger('services')
User = get_user_model()


class ServiceApiViewSet(mixins.ListModelMixin,
                        mixins.RetrieveModelMixin,
                        viewsets.GenericViewSet):
    """
    Service API viewset
    """

    serializer_class = ServiceSerializer
    permission_classes = (permissions.AllowAny, )

    def get_queryset(self):
        return Service.objects.filter(is_public=True)


class ServiceSubscriptionApiViewSet(mixins.CreateModelMixin,
                                    viewsets.GenericViewSet):
    """
    ServiceSubscription API viewset
    """

    serializer_class = ServiceSubscriptionSerializer
    permission_classes = (permissions.IsAuthenticated, )

    def get_queryset(self):
        return Service.objects.filter(is_public=True)

    def subscribe(self, request, pk, *args, **kwargs):
        logger.info('subscribing')

        requesting_user = request.user
        service = Service.objects.get(id=pk)

        if service is None:
            raise Http404

        profile = None

        if is_staff(requesting_user):
            serializer = ServiceSubscriptionSerializer(data=request.data)

            if serializer.is_valid():
                user_sso_id = serializer.validated_data.get('sso_id', None)
                if user_sso_id is not None:
                    profile = Profile.objects.filter(sso_id=user_sso_id).first()
        else:
            profile = requesting_user.sso_app_profile

        if profile is None:
            raise Http404

        if profile.subscriptions.filter(service=service).count() > 0:
            return Response('Already subscribed.', status=status.HTTP_409_CONFLICT)

        _subscription, subscription_created = service.subscribe(profile, update_rev=True)

        if subscription_created:
            response = Response('Successfully subscribed.', status=status.HTTP_201_CREATED)
        else:
            return Response('Service subscription error.', status=status.HTTP_500_INTERNAL_SERVER_ERROR)

        received_jwt = get_request_jwt(request)
        from_browser_as_same_user = received_jwt is not None

        if from_browser_as_same_user:  # From browser as same user
            logger.info(
                'From browser as same user "{}"'.format(requesting_user))
            logger.info('updating JWT for {}'.format(requesting_user))

            renew_response_jwt(received_jwt, requesting_user, request, response)

        return response

    def unsubscribe(self, request, pk, *args, **kwargs):
        logger.info('unsubscribing')

        requesting_user = request.user
        service = Service.objects.get(id=pk)

        if service is None:
            raise Http404

        profile = None

        if is_staff(requesting_user):
            serializer = ServiceSubscriptionSerializer(data=request.data)

            if serializer.is_valid():
                user_sso_id = serializer.validated_data.get('sso_id', None)
                if user_sso_id is not None:
                    profile = Profile.objects.filter(sso_id=user_sso_id).first()
        else:
            profile = requesting_user.sso_app_profile

        if profile is None:
            raise Http404

        logger.info('"{}" unsubscribing from "{}"'.format(profile, service))

        _subscription, subscription_disabled = service.unsubscribe(profile, update_rev=True)

        if subscription_disabled:
            response = Response('Successfully unsubscribed.', status=status.HTTP_200_OK)
            received_jwt = get_request_jwt(request)
            from_browser_as_same_user = received_jwt is not None

            if from_browser_as_same_user:  # From browser as same user
                logger.info(
                    'From browser as same user "{}"'.format(requesting_user))
                logger.info('updating JWT for {}'.format(requesting_user))

                renew_response_jwt(received_jwt, requesting_user, request, response)

            return response
        else:
            response = Response('Can not unsubsribe.', status=status.HTTP_400_BAD_REQUEST)

            return response


class SubscriptionApiViewSet(mixins.ListModelMixin,
                             mixins.RetrieveModelMixin,
                             viewsets.GenericViewSet):
    """
    Subscription API viewset
    """

    serializer_class = SubscriptionSerializer
    permission_classes = (permissions.IsAuthenticated,)

    def get_queryset(self):
        requesting_user = self.request.user

        if is_authenticated(requesting_user):
            if is_staff(requesting_user):
                return Subscription.objects.all()
            else:
                return Subscription.objects.filter(profile=requesting_user.sso_app_profile)
        else:
            return Subscription.objects.none()



# Generated by Django 2.2.11 on 2020-05-19 15:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_sso_app', '0004_django_user_fields_in_profile__20200417_1757'),
    ]

    operations = [
        migrations.AddField(
            model_name='profile',
            name='alias',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='alias'),
        ),
        migrations.AddField(
            model_name='profile',
            name='completed_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='profile completion date'),
        ),
        migrations.AddField(
            model_name='profile',
            name='meta',
            field=models.TextField(blank=True, null=True, verbose_name='meta'),
        ),
        migrations.AlterField(
            model_name='service',
            name='service_url',
            field=models.CharField(db_index=True, max_length=255, unique=True, verbose_name='service url'),
        ),
    ]



import logging

from django.contrib.auth import get_user_model
from django.db.models.signals import m2m_changed
from django.db.models.signals import pre_delete, pre_save
from django.dispatch import receiver

from ....permissions import is_django_staff
from ...devices.models import Device
from ...groups.models import Group
from ...profiles.models import Profile
from ..utils import delete_apigw_consumer, create_apigw_consumer_jwt, delete_apigw_consumer_jwt, \
                    create_apigw_consumer_acl, delete_apigw_consumer_acl, get_profile_apigw_consumer_id

logger = logging.getLogger('django_sso_app.core.apps.api_gateway')

User = get_user_model()


# device

@receiver(pre_save, sender=Device)
def create_device_apigw_jwt(sender, instance, **kwargs):
    device = instance
    profile = device.profile
    user = profile.user
    creating = instance._state.adding

    if creating and not is_django_staff(user):
        logger.info('Creating apigw JWT for Device "{}"'.format(device))

        apigateway_consumer_id = get_profile_apigw_consumer_id(profile)
        status_code, response = create_apigw_consumer_jwt(apigateway_consumer_id)

        if status_code != 201:
            # delete_apigw_consumer(username)
            logger.error(
                'Error ({}) creating apigw consumer JWT, "{}"'.format(
                    status_code, response))
            raise Exception(
                'Error ({}) creating apigw consumer jwt'.format(status_code))

        device.apigw_jwt_id = response.get('id')
        device.apigw_jwt_key = response.get('key')
        device.apigw_jwt_secret = response.get('secret')


@receiver(pre_delete, sender=Device)
def delete_device_api_jwt(sender, instance, **kwargs):
    device = instance
    logger.info('Deleting apigw JWT for Device "{}", jwt_id: "{}"'.format(device, device.apigw_jwt_id))

    apigateway_consumer_id = get_profile_apigw_consumer_id(device.profile)
    status_code, response = delete_apigw_consumer_jwt(apigateway_consumer_id, device.apigw_jwt_id)

    logger.debug('Kong JWT deleted ({0}), {1}'.format(status_code, response))
    if status_code >= 300:
        if status_code != 404:
            # delete_apigw_consumer(username)
            logger.error('Error ({}) Deleting apigw JWT for Device "{}", "{}"'.format(status_code, device, response))

            raise Exception(
                "Error deleting apigw consumer jwt, {}".format(status_code))


# profile

@receiver(pre_delete, sender=Profile)
def delete_apigw_profile_consumer(sender, instance, **kwargs):
    profile = instance
    user = profile.user

    if not is_django_staff(user):
        logger.info('Profile "{}" will be deleted, removing api gateway consumer'.format(profile))

        apigateway_consumer_id = get_profile_apigw_consumer_id(profile)
        status_code, response = delete_apigw_consumer(apigateway_consumer_id)
        logger.debug('Api gateway consumer deletion: ({0}), {1}'.format(status_code, response))

        if status_code >= 300:
            if status_code != 404:
                # delete_apigw_consumer(username)
                logger.error(
                    'Error ({0}) Deleting apigw consumer for User {1}, {2}'.format(
                        status_code, profile, response))
                raise Exception(
                    "Error deleting apigw consumer, {}".format(status_code))


# group

@receiver(m2m_changed, sender=Profile.groups.through)
def api_gateway_signal_handler_when_profile_is_added_to_or_removed_from_group(action, instance, pk_set, **kwargs):

    profile = instance
    user = profile.user

    if not is_django_staff(user):
        is_loaddata = getattr(user, '__dssoa__loaddata', False)
        #is_apigateway_update = getattr(user, '__dssoa__apigateway_update', False)

        apigateway_consumer_id = get_profile_apigw_consumer_id(profile)
        logger.debug('apigateway_consumer_id on m2m_changed "{}"'.format(apigateway_consumer_id))

        if action == 'pre_add':

            # do not create acl groups for signing up users
            if user.last_login is None:
                logger.info('do not create acl groups while user never logged in')
                return

            for pk in pk_set:
                group = Group.objects.get(id=pk)

                logger.info('creating api gateway acl group "{}" for profile "{}"'.format(group, profile))

                apigateway_consumer_id = get_profile_apigw_consumer_id(profile)
                status_code, resp = create_apigw_consumer_acl(apigateway_consumer_id, group.name)

                if status_code < 400:
                    logger.debug('api gateway acl group created {}, {}'.format(status_code, resp))

                    if not is_loaddata:  # or is_apigateway_update):
                        profile.update_rev(True)
                elif status_code == 409:
                    logger.debug('acl group "{} already created'.format(group.name))
                else:
                    logger.exception('error ({}) creating api gateway acl group'.format(status_code))

        elif action == 'pre_remove':
            for pk in pk_set:
                group = Group.objects.get(id=pk)
                logger.info('deleting apigw acl group "{}" for "{}"'.format(group, profile))

                apigateway_consumer_id = get_profile_apigw_consumer_id(profile)
                status_code, resp = delete_apigw_consumer_acl(apigateway_consumer_id, group.name)

                if status_code < 400:
                    logger.debug('"{}" DELETED, updating profile sso_rev'.format(group))

                    if not is_loaddata:
                        profile.update_rev(True)
                elif status_code == 404:
                    logger.debug('acl group "{} already deleted'.format(group.name))
                else:
                    logger.exception('error ({}) deleting api gateway acl group'.format(status_code))



from django.contrib import admin

from treebeard.admin import TreeAdmin
from treebeard.forms import movenodeform_factory

from .models import Grouping


class GroupingAdmin(TreeAdmin):
    form = movenodeform_factory(Grouping)
    list_display = ('name', 'type', 'group')
    search_fields = ('name', 'type', 'group__name')


admin.site.register(Grouping, GroupingAdmin)



from django.contrib.auth import get_user_model
from django.views.generic import TemplateView

from . import app_settings

User = get_user_model()


class WebpackBuiltTemplateViewMixin(TemplateView):
    """
    Uses built template view when DJANGO_SSO_APP_BACKEND_HAS_CUSTOM_FRONTEND_APP
    """

    def get_template_names(self):
        if app_settings.BACKEND_HAS_CUSTOM_FRONTEND_APP:
            original_template_path = getattr(self, 'template_name')
            relative_template_path = original_template_path

            return [relative_template_path]
        else:
            return [getattr(self, 'template_name')]

    def get_context_data(self, *args, **kwargs):
        context = super(WebpackBuiltTemplateViewMixin, self).get_context_data(*args, **kwargs)
        return context



# Boostable functions
#
# import pyximport
# pyximport.install()

import random
import hashlib
from string import ascii_uppercase, digits
from urllib.parse import urlparse


def get_random_string(n: int) -> str:
    """
    Returns n characters random string
    :param n:
    :return:  str
    """
    return ''.join(random.choice(ascii_uppercase + digits) for _ in range(n))


def get_url_host(url: str) -> str:
    """
    Returns given url host base path
    :param url:
    :return: str
    """
    parsed_url = urlparse(url)

    if parsed_url.scheme is not None and parsed_url.hostname is not None:
        return_url = parsed_url.scheme + '://' + parsed_url.hostname

        if parsed_url.port is not None:
            return_url += ':' + str(parsed_url.port)

        return return_url


def lists_differs(a: list, b: list) -> bool:
    """
    Checks if a and b differs
    :param a:
    :param b:
    :return:
    """
    a.sort()
    b.sort()

    return a != b


def get_numeric_hash(s: str,
                     HASH_LENGTH: int = 2,
                     ENCODING: str = 'utf-8') -> str:
    """
    Returns HASH_LENGTH string hash [0-1]*
    :param s:
    :param HASH_LENGTH:
    :param ENCODING:
    :return:
    """
    num = str(int(hashlib.sha256(s.encode(ENCODING)).hexdigest(), 16) % 10**HASH_LENGTH)

    if len(num) == 1:
        num = '0' + num

    return num



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class UsersConfig(AppConfig):
    name = "django_sso_app.backend.users"
    verbose_name = _("Users")

    def ready(self):
        try:
            from . import signals  # noqa F401
        except ImportError:
            pass



import logging

# import pyximport
# pyximport.install()

from django.db import transaction

from ...functions import lists_differs
from ... import app_settings

logger = logging.getLogger('django_sso_app.core.apps.groups')


@transaction.atomic
def set_default_user_groups(user):
    """
    Update User model groups.
    Checks for DJANGO_SSO_APP_DEFAULT_USER_GROUPS +
    """
    logger.debug('set_default_user_groups')

    setattr(user, '__dssoa__updating_default_groups', True)

    previous_groups_list = list(user.groups.order_by('name').values_list('name', flat=True))
    actual_groups_list = list(set(previous_groups_list + app_settings.DEFAULT_USER_GROUPS))

    if lists_differs(previous_groups_list, actual_groups_list):
        logger.info('Must update default user groups for "{}"'.format(user))
        logger.info('From {} to {}'.format(previous_groups_list, actual_groups_list))

        # try to add user to new groups
        for group_name in actual_groups_list:
            if group_name not in previous_groups_list:
                user.add_to_group(group_name)

        # try to remove user from removed groups
        for group_name in previous_groups_list:
            if group_name not in actual_groups_list:
                user.remove_from_group(group_name)

    setattr(user, '__dssoa__updating_default_groups', False)


@transaction.atomic
def set_default_profile_groups(profile):
    logger.debug('set_default_profile_groups')

    setattr(profile.user, '__dssoa__updating_default_groups', True)

    previous_groups_list = list(profile.groups.order_by('name').values_list('name', flat=True))
    actual_groups_list = list(set(previous_groups_list +
                                  app_settings.DEFAULT_PROFILE_GROUPS))

    if profile.is_incomplete:
        if 'incomplete' not in actual_groups_list:
            logger.debug('Adding to incomplete group')
            actual_groups_list.append('incomplete')
    else:
        if 'incomplete' in actual_groups_list:
            logger.debug('Removing from incomplete group')
            actual_groups_list.remove('incomplete')

    if lists_differs(previous_groups_list, actual_groups_list):
        logger.info('Must set default profile groups for "{}"'.format(profile))
        logger.info('From {} to {}'.format(previous_groups_list, actual_groups_list))

        # try to add profile to new groups
        for group_name in actual_groups_list:
            if group_name not in previous_groups_list:
                profile.add_to_group(group_name)

        # try to remove profile from removed groups
        for group_name in previous_groups_list:
            if group_name not in actual_groups_list:
                profile.remove_from_group(group_name)

    setattr(profile.user, '__dssoa__updating_default_groups', False)



import json

import jwt
from django.urls import reverse
from rest_framework import status

from ...users.tests.test_views import UserTestCase
from ..models import Service


class TestSubscriptions(UserTestCase):

    def test_user_can_list_subscriptions(self):
        new_service = self._get_new_service()
        new_pass = self._get_random_pass()
        new_user = self._get_new_subscribed_user(service=new_service, password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = client.get(
            reverse('django_sso_app_service:rest-subscription-list'),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response2.data['results']), 1)
        self.assertEqual(response2.data['results'][0].get('service').get('service_url'), new_service.service_url)

    def test_user_can_retrieve_subscription(self):
        new_service = self._get_new_service()
        new_pass = self._get_random_pass()
        new_user = self._get_new_subscribed_user(service=new_service, password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        new_service_subscription = new_user.sso_app_profile.subscriptions.first()

        response2 = client.get(
            reverse('django_sso_app_service:rest-subscription-detail', args=(new_service_subscription.id,)),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(response2.data.get('is_unsubscribed'), False)


class TestServices(UserTestCase):

    def test_user_can_list_services(self):
        new_service = self._get_new_service()
        new_pass = self._get_random_pass()
        new_user = self._get_new_subscribed_user(service=new_service, password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = client.get(
            reverse('django_sso_app_service:rest-list'),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response2.data['results']), Service.objects.count())

    def test_user_can_retrieve_service(self):
        new_service = self._get_new_service()
        new_pass = self._get_random_pass()
        new_user = self._get_new_subscribed_user(service=new_service, password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = client.get(
            reverse('django_sso_app_service:rest-detail', args=(new_service.id, )),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(response2.data.get('service_url'), new_service.service_url)

    def test_user_can_not_subscribe_profile_to_service_twice(self):
        new_service = self._get_new_service()
        new_pass = self._get_random_pass()
        new_user = self._get_new_subscribed_user(service=new_service, password=new_pass)

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = client.post(
            reverse('django_sso_app_service:rest-subscription', args=(new_service.id, )),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_409_CONFLICT)
        self.assertEqual(response2.data, 'Already subscribed.')

    def test_user_can_subscribe_to_service_with_rev_incremented_and_devices_deleted(self):
        new_service = self._get_new_service()
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)

        new_user_profile = new_user.sso_app_profile
        first_rev = new_user_profile.sso_rev

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        first_devices = new_user_profile.devices.all()
        first_devices_ids = [x.id for x in first_devices]

        response2 = client.post(
            reverse('django_sso_app_service:rest-subscription', args=(new_service.id, )),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_201_CREATED)

        new_user_profile.refresh_from_db()

        self.assertEqual(response2.data, 'Successfully subscribed.')

        new_user_profile.refresh_from_db()
        second_rev = new_user_profile.sso_rev

        self.assertEqual(second_rev, first_rev+1, 'rev not incremented')

        response3 = client.get(
            reverse('django_sso_app_service:rest-list'),
            content_type='application/json'
        )

        self.assertEqual(response3.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response3.data['results']), Service.objects.count())

        jwt_cookie2 = self._get_response_jwt_cookie(response)

        unverified_payload2 = jwt.decode(jwt_cookie2, None, False)

        self.assertEqual(unverified_payload2.get('sso_rev'), second_rev, 'cookie rev NOT changed after login')

        second_devices = new_user_profile.devices.all()
        second_devices_ids = [x.id for x in second_devices]

        self.assertEqual(len(second_devices), 1)
        self.assertNotEqual(first_devices_ids, second_devices_ids)

    def test_user_can_unsubscribe_profile_from_service_with_rev_incremented(self):
        new_service = self._get_new_service()
        new_pass = self._get_random_pass()
        new_user = self._get_new_subscribed_user(service=new_service, password=new_pass)

        new_user_profile = new_user.sso_app_profile
        first_rev = new_user_profile.sso_rev

        client = self._get_client()
        response = client.post(
            reverse('rest_login'),
            data=json.dumps(self._get_login_object(new_user.email, new_pass)),
            content_type='application/json'
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        response2 = client.post(
            reverse('django_sso_app_service:rest-unsubscription', args=(new_service.id, )),
            data=json.dumps({
                'password': new_pass
            }),
            content_type='application/json'
        )

        self.assertEqual(response2.status_code, status.HTTP_200_OK)
        self.assertEqual(response2.data, 'Successfully unsubscribed.')

        new_user_profile.refresh_from_db()
        second_rev = new_user_profile.sso_rev

        self.assertEqual(second_rev, first_rev+1, 'rev not incremented on service unsubscription')

        jwt_cookie = self._get_response_jwt_cookie(response2)
        self.assertIsNotNone(jwt_cookie, 'no new jwt received')

        unverified_payload = jwt.decode(jwt_cookie, None, False)

        self.assertEqual(unverified_payload.get('sso_rev'), second_rev, 'rev NOT updated in JWT')

    def test_staff_user_can_subscribe_profile_to_service_with_profile_rev_incremented(self):
        new_service = self._get_new_service()
        new_pass = self._get_random_pass()
        new_user = self._get_new_user(password=new_pass)
        new_user_profile = new_user.sso_app_profile

        new_staff_user = self._get_new_staff_user()

        first_rev = new_user_profile.sso_rev

        client = self._get_client()
        response = client.post(
            reverse('django_sso_app_service:rest-subscription', args=(new_service.id, )),
            data=json.dumps({'sso_id': str(new_user_profile.sso_id)}),
            **self._get_new_api_token_headers(new_staff_user)
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data, 'Successfully subscribed.')

        new_user_profile.refresh_from_db()
        second_rev = new_user_profile.sso_rev

        self.assertEqual(second_rev, first_rev+1, 'rev not incremented on service subscription')




from django.contrib import admin
from django.contrib.flatpages.admin import FlatPageAdmin
from django.contrib.flatpages.models import FlatPage
from django.utils.translation import gettext_lazy as _


# Define a new FlatPageAdmin - https://docs.djangoproject.com/en/2.2/ref/contrib/flatpages/
class FlatPageAdmin(FlatPageAdmin):
    fieldsets = (
        (None, {'fields': ('url', 'title', 'content', 'sites')}),
        (_('Advanced options'), {
            'classes': ('collapse',),
            'fields': (
                # 'enable_comments',
                'registration_required',
                'template_name',
            ),
        }),
    )

# Re-register FlatPageAdmin
admin.site.unregister(FlatPage)
admin.site.register(FlatPage, FlatPageAdmin)



# -*- coding: utf-8 -*-
"""
    Setup file for django_sso_app.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 3.2.1.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
import sys

from pkg_resources import VersionConflict, require
from setuptools import setup

try:
    require('setuptools>=38.3')
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)


if __name__ == "__main__":
    setup(use_pyscaffold=True)



from django.utils.deprecation import MiddlewareMixin


class XForwardedForMiddleware(MiddlewareMixin):

    def process_request(self, request):
        if 'HTTP_X_FORWARDED_FOR' in request.META:
            request.META['REMOTE_ADDR'] = request.META['HTTP_X_FORWARDED_FOR'].split(",")[0].strip()

        return None



import logging

from allauth.account.auth_backends import AuthenticationBackend as allauth_AuthenticationBackend

from ...utils import check_user_can_login
from ...permissions import is_authenticated

logger = logging.getLogger('django_sso_app.core.authentication.backends')


class DjangoSsoAppLoginAuthenticationBackend(allauth_AuthenticationBackend):
    def authenticate(self, request, **credentials):
        user = super(DjangoSsoAppLoginAuthenticationBackend, self).authenticate(request, **credentials)

        if is_authenticated(request.user):
            check_user_can_login(request.user,
                                 skip_profile_completion_checks=True, skip_service_subscription_checks=True)

        return user



import logging

from django.utils.translation import activate as activate_translation
from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers, exceptions
from rest_framework.exceptions import ValidationError

from allauth.account import app_settings as allauth_settings
from allauth.utils import get_username_max_length

from ..apps.emails.models import EmailAddress
from ..apps.groups.models import Group
from ..forms import LoginForm, SignupForm
from .. import app_settings

logger = logging.getLogger('django_sso_app.core.api')


class LoginSerializer(serializers.Serializer):
    form_class = LoginForm

    login = serializers.CharField()
    password = serializers.CharField(style={'input_type': 'password'})
    fingerprint = serializers.CharField(required=False, allow_null=True)

    """
    def validate_fingerprint(self, fingerprint):
        if fingerprint is None:
            msg = _('Must include "fingerprint"')
            raise exceptions.ValidationError(msg)

        return fingerprint
    """


class BaseSignupSerializer(serializers.Serializer):
    form_class = SignupForm

    username = serializers.CharField(
        max_length=get_username_max_length(),
        min_length=allauth_settings.USERNAME_MIN_LENGTH,
        required=allauth_settings.USERNAME_REQUIRED
    )
    email = serializers.EmailField(required=allauth_settings.EMAIL_REQUIRED)

    password1 = serializers.CharField(write_only=True)
    password2 = serializers.CharField(write_only=True)

    referer = serializers.CharField(required=False)

    fingerprint = serializers.CharField(required=False, allow_null=True)

    groups = serializers.ListField(child=serializers.CharField(), required=False)


if app_settings.BACKEND_SIGNUP_MUST_FILL_PROFILE:
    class SignupSerializer(BaseSignupSerializer):
        first_name = serializers.CharField(required='first_name' in app_settings.REQUIRED_PROFILE_FIELDS)
        last_name = serializers.CharField(required='last_name' in app_settings.REQUIRED_PROFILE_FIELDS,
                                          allow_null=True,
                                          allow_blank=True)

        description = serializers.CharField(required='description' in app_settings.REQUIRED_PROFILE_FIELDS,
                                            allow_null=True, allow_blank=True)
        picture = serializers.CharField(required='picture' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)
        birthdate = serializers.DateField(required='birthdate' in app_settings.REQUIRED_PROFILE_FIELDS,
                                          allow_null=True)

        latitude = serializers.FloatField(required='latitude' in app_settings.REQUIRED_PROFILE_FIELDS,
                                          allow_null=True)
        longitude = serializers.FloatField(required='longitude' in app_settings.REQUIRED_PROFILE_FIELDS,
                                           allow_null=True)

        country = serializers.CharField(required='country' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)
        country_of_birth = serializers.CharField(required='country_of_birth' in app_settings.REQUIRED_PROFILE_FIELDS,
                                                 allow_null=True)
        address = serializers.CharField(required='address' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)

        language = serializers.CharField(required='language' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)

        alias = serializers.CharField(required='alias' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)

        gender = serializers.CharField(required='gender' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)

        company_name = serializers.CharField(required='company_name' in app_settings.REQUIRED_PROFILE_FIELDS, allow_null=True)

else:
    class SignupSerializer(BaseSignupSerializer):
        pass


class PasswordResetSerializer(serializers.Serializer):
    email = serializers.EmailField()

    def validate_email(self, value):
        value = super(PasswordResetSerializer, self).validate_email(value)
        saved_email = EmailAddress.objects.get(email=value)
        if not saved_email.verified:
            raise ValidationError(
                _('E-mail address not verified: %(value)s'),
                code='invalid',
                params={'value': value},
            )

        user = saved_email.user
        user_language = user.language
        logger.info('Rendering PasswordResetSerializer mail for {0} with language {1}'.format(user, user_language))
        activate_translation(user_language)

        return value

"""
    def get_email_options(self):
        # Override this method to change default e-mail options

        return {
            'email_template_name': 'account/email/password_reset_key_message.txt',
            'html_email_template_name': 'account/email/password_reset_key_message.html',
            'use_https': app_settings.HTTP_PROTOCOL == 'https',
            'extra_email_context': {
                'EMAILS_DOMAIN': settings.EMAILS_DOMAIN,
                'EMAILS_SITE_NAME': settings.EMAILS_SITE_NAME
            }
        }
"""


class PasswordResetFromKeySerializer(serializers.Serializer):
    uidb36 = serializers.CharField()
    key = serializers.CharField()
    password1 = serializers.CharField()
    password2 = serializers.CharField()


class AddEmailSerializer(serializers.Serializer):
    email = serializers.EmailField()


class PasswordChangeSerializer(serializers.Serializer):
    oldpassword = serializers.CharField()
    password1 = serializers.CharField()
    password2 = serializers.CharField()


class PasswordSetSerializer(serializers.Serializer):
    password1 = serializers.CharField()
    password2 = serializers.CharField()
